/***********************************************************************
 * File: floating_point.h
 * 	(Extracted from
 * "$Id: floating_point.h,v 1.1 2008/08/09 18:31:18 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  	This C header defines the method interfaces that compares two
 *	double/long double floating points, satisfactory.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#ifndef _FLOATING_POINT_H_ /* { */
#define _FLOATING_POINT_H_
#ifdef __cplusplus
extern "C" {
#endif

/*
 * Refer URL:
 * "Comparing floating point numbers" by Bruce Dawson
 * http://www.cygnus-software.com/papers/comparingfloats/comparingfloats.htm
 */

/*
 * Compile switches:
 *	-DLITTLE_ENDIAN	:	When the architecture is the "Little endian"
 *						(ex. Intel x86), define this macro.
 *						The default architecture is "big endian".
 *	-DUSE_STRTOLD	:	When the C-compiler provides 'strtold()',
 *						define this macro.
 *	-DLONG_DOUBLE	:	When this macro is defined, it enables 'llo_xxx
 *						(Long Long Object)' functions, defines 'LONG_'
 *						as the 8 bytes (long long) integer and defines
 *						'DOUBLE_' as the 12 bytes (long double) floating
 *						points.
 *	-DNO_LONG_LONG	:	When this macro is defined, it enables 'llo_xxx
 *						(Long Long Object)' functions, defines 'LONG_'
 *						as the 4 bytes (long) integer and defines
 *						'DOUBLE_' as the 8 bytes (double) floating
 *						points.
 *	-DUnitTest		:	When this macro is defined, the API-s enables
 *						verbose mode.
 */

/*
 * include system definitions
 */
#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <float.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>


/*
 * Definitions of types, formatting and indexes for this method
 */
#if defined(LONG_DOUBLE) /* { */
	/*
	 * Definitions for 'long double'.
	 */
	#define LLO_LONG_DOUBLE_CNT 3	/* long array max count */
	#if defined(LITTLE_ENDIAN) /* { */
		/*
		 * for little endian architectures (ex. Intel-x86)
		 */
		#define HIGH_LLO_IDX 2	/* high digit index of llo */
		#define MID_LLO_IDX  1	/* mid digit index of llo */
		#define LOW_LLO_IDX  0	/* low digit index of llo */
		#define AS_HIGH_LONG(D)	((long *)&(D))[HIGH_LLO_IDX]
		#define AS_MID_LONG(D)	((long *)&(D))[MID_LLO_IDX]
		#define AS_LOW_LONG(D)	((long *)&(D))[LOW_LLO_IDX]
	#else /* } { !defined(LITTLE_ENDIAN) default architecture */
		/*
		 * for big endian architectures (ex. Motorola MC68x00)
		 */
		#define HIGH_LLO_IDX 0	/* high digit index of llo */
		#define MID_LLO_IDX  1	/* mid digit index of llo */
		#define LOW_LLO_IDX  2	/* low digit index of llo */
		#define AS_HIGH_LONG(D)	((long *)&(D))[HIGH_LLO_IDX]
		#define AS_MID_LONG(D)	((long *)&(D))[MID_LLO_IDX]
		#define AS_LOW_LONG(D)	((long *)&(D))[LOW_LLO_IDX]
	#endif /* } !defined(LITTLE_ENDIAN) */
#elif defined(NO_LONG_LONG) /* } { */
	/*
	 * for C-compilers unsupported the 'long long' 8bytes integers.
	 */
	#define LLO_CNT	2	/* long array max count */
	#if defined(LITTLE_ENDIAN) /* { */
		/*
		 * for little endian architectures (ex. Intel-x86)
		 */
		#define HIGH_LLO_IDX 1 /* high digit index of llo */
		#define LOW_LLO_IDX  0 /* low digit index of llo */
		#define AS_HIGH_LONG(D)	((long *)&(D))[HIGH_LLO_IDX]
		#define AS_LOW_LONG(D)	((long *)&(D))[LOW_LLO_IDX]
	#else /* } { !defined(LITTLE_ENDIAN) */
		  /*
		   * for big endian (default) architecture
		   * (ex. Motorola MC68x00)
		   */
		#define HIGH_LLO_IDX 0 /* high digit index of llo */
		#define LOW_LLO_IDX  1 /* low digit index of llo */
		#define AS_HIGH_LONG(D)	((long *)&(D))[HIGH_LLO_IDX]
		#define AS_LOW_LONG(D)	((long *)&(D))[LOW_LLO_IDX]
	#endif /* } !defined(LITTLE_ENDIAN) */
#else /* } { !defined(NO_LONG_LONG) */
	  /*
	   * for C-compiler supporting the "long long" 8bytes integers.
	   * (ex. GNU-C compiler)
	   */
	#define AS_LONG_LONG(D)	*((long long *)&(D))
	#define DEFAULT_max_ullps 10LL
	#define LONG_LONG_MAX	LLONG_MAX
	#define LONG_LONG_MIN	LLONG_MIN
#endif /* } !defined(NO_LONG_LONG) */


/*
 * declarations of the API prototyping
 */
/*
 * This API initializes this package
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'init_long_floating_point_lib()' is enable.
 *
 * 'max_ullps' typed as long pointer must be a pointer to long[LLO_CNT]
 * or long_[LLO_LONG_DOUBLE_CNT].
 *
 * If the 'max_ullps' isn't positive, this returns zero and sets 'errno'
 * to 'EINVAL (invalid argument)'.
 */
#if defined(LONG_DOUBLE) /* { */
	int init_long_floating_point_lib(int verbose, long *max_ullps);
#elif defined(NO_LONG_LONG) /* } { */
	int init_floating_point_lib(int verbose, long *max_ullps);
#else /* } { !(defined(NO_LONG_LONG) || defined(LONG_DOUBLE)) */
	int init_floating_point_lib(int verbose, long long max_ullps);
#endif /* } !(defined(NO_LONG_LONG) || defined(LONG_DOUBLE)) */


/*
 * This API sets the maximum unsigned long (or unsigned long long)
 * presentation significant (max_ullps).
 * 'max_ullps' can also be interpreted in terms of how many
 * representable floats we are willing to accept between A and B. 
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'set_max_long_ullps()' is enable.
 *
 * 'max_ullps' typed as long pointer must be a pointer to long[LLO_CNT]
 * or long_[LLO_LONG_DOUBLE_CNT].
 *
 * If the argument isn't unsigned, this returns zero and sets 'errno'
 * to 'EINVAL (invalid argument)'.
 */
#if defined(LONG_DOUBLE) /* { */
	int set_max_long_ullps(long *max_ullps);
#elif defined(NO_LONG_LONG) /* } { */
	int set_max_ullps(long *max_ullps);
#else /* } { !(defined(NO_LONG_LONG) || defined(LONG_DOUBLE)) */
	int set_max_ullps(long long max_ullps);
#endif /* } !(defined(NO_LONG_LONG) || defined(LONG_DOUBLE)) */


/*
 * This API returns the smallest number of 'double' or 'long double'.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'get_long_smallest()' is enable.
 */
#if defined(LONG_DOUBLE) /* { */
	long double get_long_smallest(void);
#else /* } { !defined(LONG_DOUBLE) */
	double get_smallest(void);
#endif /* } defined(LONG_DOUBLE) */


/*
 * This API returns non-zero, when the number is 'not a number (NaN)'
 * in IEEE754.  Otherwise, this returns zero.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'is_not_a_long_number()' is enable.
 */
#if defined(LONG_DOUBLE) /* { */
	int is_not_a_long_number(long double d);
#else /* } { !defined(LONG_DOUBLE) */
	int is_not_a_number(double d);
#endif /* } defined(LONG_DOUBLE) */


/*
 * This API returns non-zero, when the number is 'infinite' in IEEE754.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'is_long_infinite()' is enable.
 */
#if defined(LONG_DOUBLE) /* { */
	int is_long_infinite(long double d);
#else /* } { !defined(LONG_DOUBLE) */
	int is_infinite(double d);
#endif /* } defined(LONG_DOUBLE) */


/*
 * This API returns non-zero, when the number is negative.
 * Otherwise, this returns zero.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'is_long_negative()' is enable.
 */
#if defined(LONG_DOUBLE) /* { */
	int is_long_negative(long double d);
#else /* } { !defined(LONG_DOUBLE) */
	int is_negative(double d);
#endif /* } defined(LONG_DOUBLE) */


/*
 * This API returns non-zero, when the numbers equals.
 * Otherwise, this returns zero.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'equals_long_doubles()' is enable.
 *
 * If the 'max_ullps' is negative, it means that this API wasn't
 * initialized, yet.  This returns 0 and set 'errno' to 'EINVAL'.
 * If the number is 'not a number', this returns zero and set 'errno'
 * to 'ERANGE'.
 */
#if defined(LONG_DOUBLE) /* { */
	int equals_long_doubles(long double d1, long double d2);
#else /* } { !defined(LONG_DOUBLE) */
	int equals_doubles(double d1, double d2);
#endif /* } defined(LONG_DOUBLE) */


/*
 * This API returns the result of comparing two numbers.
 * If d1 equals d2, this returns zero.
 * If d1 is least than d2, this returns the negative.
 * Otherwise (d1 is greater than d2), this returns positive.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'cmp_long_double()' is enable.
 *
 * If the 'max_ullps' is negative, it means that this API wasn't
 * initialized, yet.  This returns 0 and set 'errno' to 'EINVAL'.
 * If the number is 'not a number', this returns zero and set 'errno'
 * to 'ERANGE'.
 */
#if defined(LONG_DOUBLE) /* { */
	int cmp_long_double(long double d1, long double d2);
#else /* } { !defined(LONG_DOUBLE) */
	int cmp_double(double d1, double d2);
#endif /* } defined(LONG_DOUBLE) */


/*
 * This API returns the minimum size of the formatted string for
 * 'double' or 'long double'.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'get_min_size_of_long_double_string_buf()' is enable.
 */
#if defined(LONG_DOUBLE) /* { */
	size_t get_min_size_of_long_double_string_buf(void);
#else /* } { !defined(LONG_DOUBLE) */
	size_t get_min_size_of_double_string_buf(void);
#endif /* } defined(LONG_DOUBLE) */


/*
 * This API returns the 'double' or 'long double' that is converted
 * a string.
 *
 * If '-DLONG_DOUBLE' is specified to the compiler,
 * 'string2long_double()' is enable.
 *
 * If the string is NULL, this returns negative infinite and set 'errno'
 * to 'EINVAL'.
 * If the string is 'not a number' in IEEE754, this returns NaN and set
 * 'errno' to 'ERANGE'.
 * If failed to convert the string, this returns zero or negative zero
 * and set 'errno' to 'ERANGE'.
 */
#if defined(LONG_DOUBLE) /* { */
	long double string2long_double(const char *s);
#else /* } { !defined(LONG_DOUBLE) */
	double string2double(const char *s);

#endif /* } defined(LONG_DOUBLE) */

#ifdef __cplusplus
}
#endif
#endif /* } !defined(_FLOATING_POINT_H_) */
