/***********************************************************************
 * File: utf8.h
 * 	(Extracted from
 *	   "$Id: utf8.h,v 1.1 2008/10/05 15:55:33 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C header defines prototyping for the UTF-8 character decoder
 *  and encoder.
 *
 * Copyright (C) 2008 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/


#ifndef _UTF8_H_ /* { */
#define _UTF8_H_
#ifdef __cplusplus
extern "C" {
#endif

/*
 * including the system definitions
 */
#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <iconv.h>
#include <errno.h>

typedef enum Sanitize_for_UTF8_e { /* the source is encoded in... */
	  Sanitize_for_UTF8_e_Zero			= 0
	, Sanitize_for_UTF8_e_EUC_JISX0213	/* "EUCJP-JISX0213" */
	, Sanitize_for_UTF8_e_EUCJP_OPEN	/* "EUCJP-OPEN" */
	, Sanitize_for_UTF8_e_EUCJP_WIN		/* "EUCJP-WIN" */
	, Sanitize_for_UTF8_e_EUCJP_MS		/* "EUCJP-MS" */
	, Sanitize_for_UTF8_e_EUCJP			/* "EUCJP" */
	, Sanitize_for_UTF8_e_SJIS_OPEN		/* "SJIS-OPEN" (-> "EUCJP-OPEN") */
	, Sanitize_for_UTF8_e_SJIS_WIN		/* "SJIS-WIN"  (-> "EUCJP_WIN") */
	, Sanitize_for_UTF8_e_SJIS			/* "SJIS"      (-> "SJIS") */
	, Sanitize_for_UTF8_e_N				/* number of the source encoding. */
	} Sanitize_for_UTF8_e;

/* translate `str_encoding' to iconv_open(3)'s encode identifier */
#define Sanitize_for_UTF8_e_To_iconv_open_id(STR_ENCODING)	(			\
			((STR_ENCODING) == Sanitize_for_UTF8_e_EUC_JISX0213)?		\
		  											"EUCJP-JISX0213"	\
			: (((STR_ENCODING) == Sanitize_for_UTF8_e_EUCJP_OPEN)?		\
													"EUCJP-OPEN"		\
			: (((STR_ENCODING) == Sanitize_for_UTF8_e_EUCJP_WIN)?		\
													"EUCJP-WIN"			\
			: (((STR_ENCODING) == Sanitize_for_UTF8_e_EUCJP_MS)?		\
													"EUCJP-MS"			\
			: (((STR_ENCODING) == Sanitize_for_UTF8_e_EUCJP)?			\
													"EUCJP"				\
			: (((STR_ENCODING) == Sanitize_for_UTF8_e_SJIS_OPEN)?		\
													"SJIS-OPEN"			\
			: (((STR_ENCODING) == Sanitize_for_UTF8_e_SJIS_WIN)?		\
													"SJIS-WIN"			\
			: (((STR_ENCODING) == Sanitize_for_UTF8_e_SJIS)?			\
													"SJIS"				\
			:										NULL				\
			))))))))

/* iconv_open(3)'s identifier of UTF-8 */
#define Sanitize_for_UTF8_iconv_open_id	"UTF-8"

/* `how' arguments are the logical-or-ed the followings...
 * (default, to_utf8() converts Yen-sign (\x5c)  to UTF-8 "\xc2\xa5 U+00A5",
 * convert tilde (0x7e) to UTF-8 "\x7e"), replace the hankaku-kana with
 * `replace_code' character, and replace the machine-depended Chinese
 * character with `replace_code'.)
 */
#define Sanitize_for_UTF8_Default_how				(0UL)
#define Sanitize_for_UTF8_Yen_sign_must_be_U005C	(1UL)
#define Sanitize_for_UTF8_Tilde_must_be_U203E		(2UL)
#define Sanitize_for_UTF8_Pass_hankaku_kana			(4UL)
#define Sanitize_for_UTF8_Pass_extra_characters		(8UL)
#define Sanitize_for_UTF8_Set_all					(					\
							  Sanitize_for_UTF8_Default_how				\
							| Sanitize_for_UTF8_Yen_sign_must_be_U005C	\
							| Sanitize_for_UTF8_Tilde_must_be_U203E		\
							| Sanitize_for_UTF8_Pass_hankaku_kana		\
							| Sanitize_for_UTF8_Pass_extra_characters	\
							)

/* this function sanitizes a string (safer UTF-8 encoding)
 * if error occurred, this returns NULL and set `errno'.
 *	EINVAL means "illegal arguments".
 */
extern unsigned char /*@null@*/	*sanitize_for_utf8(
										  Sanitize_for_UTF8_e	src_encode
										, size_t			length
										, unsigned char		*str
										, char				replace_code
										, unsigned long		how
										);

/* convert the EUC-JP or SJIS encoded string to UTF-8 encoded.
 * if error occurred, this returns NULL and set `errno'.
 *	EINVAL means "illegal arguments".
 *	ENOMEM means "insufficient memory".
 *	EPERM means "system not support EUC-JP, SJIS or UTF-8 encode".
 *	EILSEQ means "`inbuf' contains illegal character".
 */
extern unsigned char /*@null@*/	*to_utf8(Sanitize_for_UTF8_e	inbuf_encode
										, const unsigned char	*inbuf
										, size_t				inbuf_length
										, char					replace_code
										, unsigned char			*outbuf
										, size_t				outbuf_size
										, char					padding_code
										, unsigned long			how
										);

/* convert the EUC-JP or SJIS encoded string to UTF-8 encoded.
 * if error occurred, this returns NULL and set `errno'.
 *	EINVAL means "illegal arguments".
 *	ENOMEM means "insufficient memory".
 *	EPERM means "system not support EUC-JP, SJIS or UTF-8 encode".
 *	EILSEQ means "`inbuf' contains illegal character".
 */
extern unsigned char /*@null@*/	*from_utf8(
									  Sanitize_for_UTF8_e	outbuf_encode
									, unsigned char			*inbuf
									, size_t				inbuf_length
									, unsigned char			*outbuf
									, size_t				outbuf_size
									, char					padding_code
									);

#ifdef __cplusplus
}
#endif
#endif /* } (!defined(_UTF8_H_) */
