/***********************************************************************
 * File: aggregate.c
 * 	(Extracted from
 *	   "$Id: aggregate.c,v 1.1 2009/05/03 13:22:09 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source file implements `aggregate()' API.
 *
 * Copyright (C) 2009 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "aggregate.h" /* definitions of `aggregate' API-s */

/* definition of signs for `Number_t' */
typedef enum sign_e {
	  Sign_e_positive	/* sign is positive. */
	, Sign_e_negative	/* sign is negative. */
}	Sign_e;

/* type definition of the number for processing */
typedef struct number_t {
	const char	*ip;	/* pointer to integer part */
	const char	*dp;	/* pointer to decimal point */
	const char	*ep;	/* pointer to end of numeral string */
	Sign_e	sign;	/* sign of number */
} Number_t;

/* the Number_t structure `x' means zero or not? */
static int is_zero(Number_t	*x)
{
	int	result	= (((x->ip == x->dp) && x->ip == x->ep)? 1: 0);

	return	result;
}

/* make a Number_t structure as the numeral string */
/*@null@*/ static Number_t *make_number(			const char	*numeral_str
										, /*@out@*/	Number_t	*number
										)
{
	const char	*p = NULL;

	/* test: arguments are invalid? */
	if ((number == NULL) || numeral_str == NULL) {
		return	NULL; /* Error: invalid arguments */
	}

	/* initialize `number' structure */
	memset(number, (int)'\0', sizeof(*number));

	/* set pointers to `decimal point' and 'end of fraction part' */
	p	= numeral_str + strlen(numeral_str);
	if ((number->dp = strchr(numeral_str, '.')) == NULL) {
		while ((numeral_str < p) && *(p - 1) == ' ') {	--p;	}
		number->dp	= (char *)p;
	}
	while ((number->dp < p) && *(p - 1) == ' ') {	--p;	}
	while ((number->dp < p) && *(p - 1) == '0') {	--p;	}
	number->ep	= ((number->dp == (p - 1))? number->dp:	(char *)p);

	/* test: fraction part is filled with digits? */
	if (number->dp != number->ep) {
		for (p = number->dp + 1; p < number->ep; ++p) {
			if (!isdigit(*p)) {
				return	NULL; /* Error: invalid arguments */
	}	}	}

	/* set sign and pointer to `start of integer part' */
	p	= numeral_str;
	while ((p < number->dp) && *p == ' ') {	++p;	}
	number->sign	= (((p == number->dp) || *p != '-')?	Sign_e_positive
						:									Sign_e_negative
						);
	if ((*p == '+') || *p == '-') {	++p;	}
	if (p == number->dp) {
		number->ip	= (char *)p;
	}
	else {
		if ((*p != ' ') && !isdigit(*p)) {
			return	NULL; /* Error: invalid arguments */
		}
		while ((p < number->dp) && *p == ' ') {	++p;	}
		while ((p < number->dp) && *p == '0') {	++p;	}
		number->ip	= (char *)p;
	}

	/* test: integer part is filled with digits? */
	if (number->ip != number->dp) {
		for (p = number->ip; p < number->dp; ++p) {
			if (!isdigit(*p)) {
				return NULL; /* Error: invalid arguments */
	}	}	}

	/* test: numeral means zero? */
	if (is_zero(number) != 0) {
		/* zero has positive sign */
		number->sign	= Sign_e_positive;
	}

	return	number;	/* normal end */
}

/* compare two Number_t structures 'a' and 'b'
 * if `a' is less than `b', this returns -1.
 * if `a' equals `b', this returns zero.
 * if `a' is greater than `b', this returns 1.
 */
static int compare(Number_t	*a, Number_t	*b)
{
	size_t		a_len, b_len, limit;
	int			result;

	/* compare sign */
	if ((a->sign == Sign_e_negative) && b->sign == Sign_e_positive) {
			return	-1;	/* `a' is less than `b' */
	}
	else if ((b->sign == Sign_e_negative) && a->sign == Sign_e_positive) {
			return	1;	/* `a' is greater than `b' */
	}

	/* compare integer parts */
	a_len = (size_t)(a->dp - a->ip); b_len = (size_t)(b->dp - b->ip);
	if (a_len > b_len) {
		/* `a' is greater/less than `b' */
		return	((a->sign == Sign_e_positive)?	1:	-1);
	}
	else if (a_len < b_len) {
		/* `a' is less/greater than `b' */
		return	((a->sign == Sign_e_positive)?	-1:	1);
	}
	else if (a_len > 0) {
		if ((result = memcmp(a->ip, b->ip, a_len)) > 0) {
			/* `a' is greater/less than `b' */
			return	((a->sign == Sign_e_positive)?	1:	-1);
		}
		else if (result < 0) {
			/* `a' is less/greater than `b' */
			return	((a->sign == Sign_e_positive)?	-1:	1);
		}
	}

	/* compare fraction parts */
	a_len = (size_t)(a->ep - a->dp); b_len = (size_t)(b->ep - b->dp);
	limit	= ((a_len < b_len)? a_len: b_len);
	if (limit > 0) {
		if ((result = memcmp(a->dp , b->dp, limit)) > 0) {
			/* `a' is greater/less than `b' */
			return	((a->sign == Sign_e_positive)?	1:	-1);
		}
		else if (result < 0) {
			/* `a' is less/greater than `b' */
			return	((a->sign == Sign_e_positive)?	-1:	1);
		}
	}
	if (a_len > b_len) {
		/* `a' is greater/less than `b' */
		return	((a->sign == Sign_e_positive)?	1:	-1);
	}
	else if (a_len < b_len) {
		/* `a' is less/greater than `b' */
		return	((a->sign == Sign_e_positive)?	-1:	1);
	}

	return	0;	/* `a' equals `b' */
}

/* convert a digit code to integer */
static int digit2int(const char	*p)
{
	char	buf[2];

	buf[0] = *p; buf[1] = '\0'; return atoi(buf);
}

/* convert an integer to a digit code */
static char int2digit(int d)
{
	static char	digits[] =	{ '0', '1', '2', '3', '4'
							, '5', '6', '7', '8', '9'
							};
	return	digits[d];
}

/* add two Number_t structure `x' and `y'
 * signs of `a' and `b' must be `Sign_e_positive'
 */
/*@null@*/ static char *add(Number_t	*x
							, Number_t	*y
							, char		*result
							, size_t	result_size
							)
{
	const char	*xp, *yp;
	size_t		i, x_len, y_len, move_len;
	char		*rp;
	int			carry;

	/* set end of string */
	rp = result + result_size; *rp = '\0';

	/* process fraction parts */
	if ((x->ep != x->dp) || y->ep != y->dp) {
		if (x->ep == x->dp) {
			/* `x' has no fraction parts, `y' has it */
			if ((move_len = (size_t)(y->ep - y->dp)) > (result_size - 1)) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			rp	-= move_len;
			memcpy(rp, y->dp, move_len);
			xp	= x->dp; yp = y->dp; carry = 0;
		}
		else if (y->ep == y->dp) {
			/* `y' has no fraction part, `x' has it */
			if ((move_len = (size_t)(x->ep - x->dp)) > (result_size - 1)) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			rp	-= move_len;
			memcpy(rp, x->dp, move_len);
			xp	= x->dp; yp = y->dp; carry = 0;
		}
		else {
			/* adding the fraction digits */
			x_len = (size_t)(x->ep - x->dp)
			; y_len = (size_t)(y->ep - y->dp)
			;
			if (x_len > y_len) {
				/* `x' fraction part is longer than `y's */
				if ((move_len = x_len - y_len) > (result_size - 1)) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				rp	-= move_len;
				xp	= x->dp + (x_len - move_len);
				yp	= y->dp + y_len;
				memcpy(rp, xp, move_len);
				move_len	= y_len;
			}
			else if (x_len < y_len) {
				/* `y' fraction part is longer than `x's */
				if ((move_len = y_len - x_len) > (result_size - 1)) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				rp	-= move_len;
				xp	= x->dp + x_len;
				yp	= y->dp + (y_len - move_len);
				memcpy(rp, yp, move_len);
				move_len	= x_len;
			}
			else {
				/* 'x' fraction part is such length as `y's */
				xp	= x->dp + x_len;
				yp	= y->dp + y_len;
				move_len	= x_len;
			}

			for (i = 1, carry = 0; i < move_len; ++i) {
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				--xp; --yp;
				carry	+= (digit2int(xp) + digit2int(yp));
				*rp	= 	int2digit(carry % 10);
				carry	/= 10;
			}

			/* set decimal point */
			if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			--xp; --yp; *rp	= *xp;
		}
	}
	else {
		/* `x' and `y' has no fraction part */
		xp = x->ep; yp = y->ep; carry = 0;
	}

	/* process integer parts */
	x_len = (size_t)(xp - x->ip); y_len = (size_t)(yp - y->ip);
	move_len	= ((x_len < y_len)?	x_len: y_len);
	for (i = 0; i < move_len; ++i) {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		--xp; --yp;
		carry	+= (digit2int(xp) + digit2int(yp));
		*rp		= int2digit(carry % 10);
		carry	/= 10;
	}
	if (y_len > x_len) {
		/* 'y's integer part is longer than `x's */
		xp	= yp; /* process `y's integer part */
		move_len	= y_len - move_len;
	}
	else {
		/* 'x's integer part is longer than `y's */
		move_len	= x_len - move_len;
	}
	for (i = 0; i < move_len; ++i) {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		--xp; --yp;
		carry	+= digit2int(xp);
		*rp		= int2digit(carry % 10);
		carry	/= 10;
	}
	if (carry != 0) {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		*rp		= int2digit(carry);
		carry	= 0;
	}

	/* set zero ('0'), if integer part is empty */
	if ((*rp == '\0') || *rp == '.') {
		if (--rp < result) {
			/* Error: insufficient size of result */
			errno = ERANGE; return NULL;
		}
		*rp	= int2digit(0);
	}

	/* move the result string to the left */
	if (rp > result) {	strcpy(result, rp);	}

	return	result; /* return the result */
}

/* subtract two Number_t structures `a' and `b'
 * signs of `a' and `b' must be Sign_e_positive
 * `a' should be greater than `b', or `a' should equals `b'
 */
/*@null@*/ static char *sub(Number_t	*x
							, Number_t	*y
							, char		*result
							, size_t	result_size
							)
{
	const char	*xp, *yp;
	size_t		x_len, y_len, len, i;
	char		*rp;
	int			bollow;

	rp = result + result_size; *rp = '\0';

	/* process the fraction part */
	if ((x->dp != x->ep) || y->dp != y->ep) {
		/* `x' or `y' has fraction parts */
		x_len = (size_t)(x->ep - x->dp); y_len = (size_t)(y->ep - y->dp);
		if (x_len > y_len) {
			/* `x's fraction part is longer than `y's */
			len	= x_len - y_len; xp = x->ep - len; x_len -= len; yp = y->ep;
			if ((rp -= len) < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			memcpy(rp, xp, len);
			bollow = 0;
		}
		else if (x_len < y_len) {
			/* `y's fraction part is longer than `x's */
			len	= y_len - x_len; xp = x->ep; yp = y->ep; y_len -= len;
			if (x_len == 0) {
				/* we don't process the decimal point digit on `y' */
				--len;
			}
			for (i = 0, bollow = 10; i < len; bollow = 9, ++i) {
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				*rp	= int2digit(bollow - digit2int(--yp));
			}
			if (x_len == 0) {
				/* set the decimal point */
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				*rp	= *(--yp);
			}
			bollow	= 1;
		}
		else {
			/* `x' is such long as `y's */
			xp = x->ep; yp = y->ep; bollow = 0;
		}

		/* subtract fraction parts of `x' and 'y' */
		if (x_len > 0) {
			len = x_len - 1;
			for (i = 0; i < len; ++i) {
				if (--rp < result) {
					/* Error: insufficient size of result */
					errno = ERANGE; return NULL;
				}
				if ((bollow = digit2int(--xp) - bollow - digit2int(--yp))
																	< 0) {
					*rp	= int2digit(10 + bollow);	bollow = 1;
				}
				else {
					*rp	= int2digit(bollow);		bollow = 0;
				}
			}
			if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
			}
			*rp	= *(--xp); --yp; /* set the decimal point */
		}
	}
	else {
		/* `x' and `y' has no fraction parts */
		xp = x->dp; yp = y->dp; bollow = 0;
	}

	/* process the integer parts */
	x_len = (size_t)(x->dp - x->ip); y_len = (size_t)(y->dp - y->ip);
	for (i = 0; i < y_len; ++i) {
		if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
		}
		if ((bollow = digit2int(--xp) - bollow - digit2int(--yp)) < 0) {
			*rp	= int2digit(10 + bollow);	bollow = 1;
		}
		else {
			*rp	= int2digit(bollow);		bollow = 0;
		}
	}
	len	= x_len - y_len;
	for (i = 0; i < len; ++i) {
		if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
		}
		if ((bollow = digit2int(--xp) - bollow) < 0) {
			*rp	= int2digit(10 + bollow);	bollow = 1;
		}
		else {
			*rp	= int2digit(bollow);		bollow = 0;
		}
	}

	/* truncate zero codes */
	while (((*rp != '\0') && *(rp + 1) != '.') && *rp == '0') {	++rp;	}

	/* set zero ('0'), if integer part is empty */
	if ((*rp == '\0') || *rp == '.') {
		if (--rp < result) {
				/* Error: insufficient size of result */
				errno = ERANGE; return NULL;
		}
		*rp	= int2digit(0);
	}

	/* move the result string to the left */
	if (rp > result) {	strcpy(result, rp);	}

	return	result;	/* return the result */
}

/* API: aggregate the numeral strings */
/*@null@*/ char *aggregate_numbers(Aggregate_e_opcode	opcode
									, const char		**numerals
									, size_t			n_numerals
									, char				*result
									, size_t			result_size
									)
{
	Number_t	a, b;
	size_t		i, j, k;
	char		*sums[2], *retval = result;

	/* test: arguments are invalid */
	if ((((((numerals == NULL)
	|| n_numerals < 1)
	|| result == NULL)
	|| result_size == 0)
	|| opcode < Aggregate_e_opcode_add)
	|| opcode >= Aggregate_e_opcode_n
	) {
		errno	= EINVAL; return	NULL; /* Error: invalid arguments */
	}

	/* allocate working buffer, and set the first numeral to it */
	if ((sums[1] = malloc(result_size)) == NULL) {
		return	NULL; /* Error: insufficient memories */
	}
	sums[1][(result_size - 1)]	= '\0';

	/* set the first numeral to `result' buffer */
	j = 0; sums[j] = result; strcpy(sums[j], numerals[0]);

	/* aggregate numerals[1] to numerals[n_numerals]... */
	for (i = 1; (retval != NULL) && i < n_numerals; ++i, j = k) {
		/* test: arguments are invalid? */
		if ((((numerals[i] == NULL)
		|| (strlen(numerals[i]) + 1) > result_size)
		|| make_number(sums[j],		&a) == NULL)
		|| make_number(numerals[i],	&b) == NULL
		) {
			/* Error: invalid arguments */
			errno = EINVAL; retval = NULL; break;
		}

		/* set index of `sums[]' as result of calculating */
		k	= ((j + 1) % (sizeof(sums) / sizeof(sums[0])));

		if (((opcode == Aggregate_e_opcode_sub) && a.sign != b.sign)
		|| ((opcode == Aggregate_e_opcode_add) && a.sign == b.sign)
		) {
			Sign_e	sign;
			size_t	result_left_size;
			char	*resultp;

			sign = a.sign; a.sign = b.sign = Sign_e_positive;
			if (sign == Sign_e_negative) {
				if (result_size == 1) {
					/* Error: insufficient size of `result' */
					errno = ERANGE; retval = NULL; break;
				}

				sums[k][0]			= '-'
				; resultp			= (sums[k] + 1)
				; result_left_size = result_size - 1
				;
			}
			else {
				resultp				= sums[k]
				; result_left_size	= result_size
				; 
			}
			retval	= add(&a, &b, resultp, result_left_size);
		}
		else if ((((opcode == Aggregate_e_opcode_sub)
					&& a.sign == Sign_e_positive)
					&& b.sign == Sign_e_positive)
		|| (((opcode == Aggregate_e_opcode_add)
					&& a.sign == Sign_e_positive)
					&& b.sign == Sign_e_negative)
		) {
			a.sign = b.sign = Sign_e_positive;
			if (compare(&a, &b) < 0) {
				if (result_size == 1) {
					/* Error: insufficient size of `result' */
					errno = ERANGE; retval = NULL; break;
				}

				sums[k][0]	= '-';
				retval	= sub(&b, &a, (sums[k] + 1), (result_size - 1));
			}
			else {
				retval	= sub(&a, &b, sums[k], result_size);
			}
		}
		else {
			a.sign = b.sign = Sign_e_positive;
			if (compare(&b, &a) < 0) {
				if (result_size == 1) {
					/* Error: insufficient size of `result' */
					errno = ERANGE; retval = NULL; break;
				}

				sums[k][0]	= '-';
				retval	= sub(&a, &b, (sums[k] + 1), (result_size - 1));
			}
			else {
				retval	= sub(&b, &a, sums[k], result_size);
			}
		}
	}

	/* if allocated buffer contains the last result,
	 * we must copy it to `result' buffer
	 */
	if ((retval != NULL) && j != 0) {
		strcpy(sums[0], sums[1]);
	}

	/* release the allocated buffer */
	memset(sums[1], (int)'\0', result_size); free(sums[1]); sums[1] = NULL;

	/* if several errors happened, this API returns NULL.
	 * otherwise, this returns address of `result' buffer
	 */
	return	retval;
}

/* API: compare two numeral strings */
int *compare_numbers(const char	*a_str, const char	*b_str, int	*result)
{
	Number_t	a, b;

	/* test: arguments are invalid
	 * convert the numeral strings to Number_t structures
	 */
	if (((result == NULL)
	|| make_number(a_str, &a) == NULL)
	|| make_number(b_str, &b) == NULL
	) {
		errno = EINVAL; return NULL; /* Error: invalid arguments */
	}

	/* return result of compare */
	*result	= compare(&a, &b); return result;
}
